/* bapi.h */
#include "frame.h"

#define BP_TYPE_VARIABLE 0x100L
#define BP_TYPE_INTEGER  0x101L
#define BP_TYPE_FLOAT    0x102L
#define BP_TYPE_ATOM     0x103L
#define BP_TYPE_COMPOUND 0x104L

#define BP_pred_ref SYM_REC_PTR

#define BP_term_ref TERM

#define BP_check_integer(A) { \
  if (!ISINT(A)){ \
      exception = illegal_arguments; \
      return BP_ERROR; \
  } \
}

#define BP_check_float(A) { \
  if (!ISFLOAT(A)){ \
      exception = illegal_arguments; \
      return BP_ERROR; \
  } \
}

#define BP_check_atom(A) { \
  if (!ISATOM(A)){ \
      exception = illegal_arguments; \
      return BP_ERROR; \
  } \
}

#define NUMERIC_VAL(op) (ISINT(op) ? (double)INTVAL(op) : floatval(op))

#define FLOAT_MINUS(op) encodefloat1(0.0-floatval(op))
#define FLOAT_LOG(op) encodefloat1(log((ISINT(op) ? (double)INTVAL(op) : floatval(op))))
#define FLOAT_LOG2(op1,op2) encodefloat1(log(NUMERIC_VAL(op2))/log(NUMERIC_VAL(op1)))
#define FLOAT_SQRT(op) encodefloat1(sqrt((ISINT(op) ? (double)INTVAL(op) : floatval(op))))
#define FLOAT_ABS(op) (ISINT(op) ? MAKEINT(abs(INTVAL(op))) : encodefloat1(fabs(floatval(op))))
#define FLOAT_EXP(op) encodefloat1(exp((ISINT(op) ? (double)INTVAL(op) : floatval(op))))
#define FLOAT_SIN(op) encodefloat1(sin((ISINT(op) ? (double)INTVAL(op) : floatval(op))))
#define FLOAT_COS(op) encodefloat1(cos((ISINT(op) ? (double)INTVAL(op) : floatval(op))))
#define FLOAT_TAN(op) encodefloat1(tan((ISINT(op) ? (double)INTVAL(op) : floatval(op))))
#define FLOAT_ATAN(op) encodefloat1(atan((ISINT(op) ? (double)INTVAL(op) : floatval(op))))
#define FLOAT_ASIN(op) encodefloat1(asin((ISINT(op) ? (double)INTVAL(op) : floatval(op))))
#define FLOAT_ACOS(op) encodefloat1(acos((ISINT(op) ? (double)INTVAL(op) : floatval(op))))
#define FLOAT_FLOAT(op) (ISINT(op) ? encodefloat1(INTVAL(op)*1.0) : op)
#define FLOAT_INTEGER(op) (ISINT(op) ? op : MAKEINT((int)(floatval(op))))
#define FLOAT_ROUND(op) (ISINT(op) ? op : MAKEINT(float_sign(floatval(op))*(BPLONG)(fabs(floatval(op))+0.5)))

#define PUSHTRAIL(val)   \
        if (((BPLONG_PTR)(val) > breg) || ((BPLONG_PTR)(val) < hbreg)) { \
        PUSHTRAILC(val,val);}

#define PUSHTRAIL_h(val)   \
        if ((BPLONG_PTR)(val) < hbreg) { \
	PUSHTRAILC(val,val);}

#define PUSHTRAIL_s(val) \
  if (((BPLONG_PTR)(val) > breg)) \
       PUSHTRAILC(val,val)

#define PUSHTRAIL_S_NONATOMIC(p,val) \
  if ((BPLONG_PTR)(p) > B) \
       PUSHTRAILC(ADDTAG3(p,TRAIL_VAL_NONATOMIC),val)

#define PUSHTRAIL_H_NONATOMIC(p,cont) \
  if ((BPLONG_PTR)p<hbreg) {\
       PUSHTRAILC(ADDTAG3(p,TRAIL_VAL_NONATOMIC),cont);}

#define PUSHTRAIL_H_BIT_VECTOR(p,cont) \
  if ((BPLONG_PTR)p<hbreg) {\
       PUSHTRAILC(ADDTAG3(p,TRAIL_BIT_VECTOR),cont);}

#define PUSHTRAIL_H_ATOMIC(p,cont) \
  if ((BPLONG_PTR)p<hbreg) {\
       PUSHTRAILC(ADDTAG3(p,TRAIL_VAL_ATOMIC),cont);}

#define PUSHTRAILC_ATOMIC(p,cont) \
       PUSHTRAILC(ADDTAG3(p,TRAIL_VAL_ATOMIC),cont);

#define PUSHTRAIL_UC_0(addr) PUSHTRAILC(addr,addr)

#define PUSHTRAILC(addr,val) {\
        if (trail_top<=trail_water_mark0){ \
                trail_top = expand_trail(trail_top,breg); \
         }; \
        *trail_top-- = (BPLONG)(val); \
        *trail_top-- = (BPLONG)(addr);}

#define POPTRAIL(t){ \
  ++t; op1 = UNTAGGED3(FOLLOW(t)); \
  ++t; *(BPLONG_PTR)op1 = FOLLOW(t); \
  }

#define UNDO_TRAILING { \
  register BPLONG op1; \
  while (trail_top != trail_top0) { \
    POPTRAIL(trail_top); \
  }}

#define IsNumberedVar(term) (((term) & TAG_MASK)==TAG_MASK)
#define NumberVar(num) (((BPLONG)(num) << 2) | TAG_MASK)

#define ARG(i,arity) *(arreg+arity-i+1)

#define NEW_VAR(op)                      \
  op = (BPLONG)heap_top;                       \
  NEW_HEAP_FREE;

#define ASSIGN_STRUCT(op1,op2)           \
  *(BPLONG_PTR)(op1) = ADDTAG(heap_top,STR);  \
  PUSHTRAIL(op1);                        \
  NEW_HEAP_NODE((BPLONG)op2);

#define ASSIGN_LIST(op)                  \
  *(BPLONG_PTR)(op) = ADDTAG(heap_top,LST);   \
  PUSHTRAIL(op);

#define MOVE_STRUCT(op1,op2)             \
  op1 = ADDTAG(heap_top,STR);               \
  NEW_HEAP_NODE((BPLONG)op2);

#define ASSIGN_VALUE(var,value)           \
  *(BPLONG_PTR)var = value;   \
  PUSHTRAIL(var)

#define ASSIGN_TRAIL_VALUE(var,value)           \
  *(BPLONG_PTR)var = (BPLONG)(value);   \
  PUSHTRAILC(var,var)

#define ASSIGN(var,value)                \
  *(BPLONG_PTR)var = value

#define LOCAL_OVERFLOW_CHECK(src) \
      if (local_top-SMALL_MARGIN < heap_top) { \
      myquit(STACK_OVERFLOW,src);\
      }

#define QUIT_IF_OVERFLOW_HEAP(margin) if (arreg-margin < heap_top){ \
     printf("Control stack overflow. \nPlease restart B-Prolog with more stack space as: bp -s xxxx (xxxx>%d)\n",stack_size);\
     exit(1); \
}

#define BUILD_VALUE(op1,lab1) \
  SWITCH_OP_VAR(op1,lab1, \
	    {if ((BPLONG_PTR)op1<heap_top) \
	       NEW_HEAP_NODE(op1); \
	    else { \
		FOLLOW(op1) = (BPLONG)heap_top; \
		PUSHTRAIL(op1); \
		NEW_HEAP_FREE;}}, \
            {NEW_HEAP_NODE(UNTAGGED_ADDR(op1));}, \
	    {NEW_HEAP_NODE(op1);})

#define GET_CAR(op) FOLLOW((BPLONG_PTR)UNTAGGED_ADDR(op))
#define GET_CDR(op) FOLLOW((BPLONG_PTR)UNTAGGED_ADDR(op)+1)
#define GET_ARG(op,n) FOLLOW((BPLONG_PTR)UNTAGGED_ADDR(op)+n)

#define LIST_NEXT(op) *(op+1)

/* tail and new_list maybe the same */
#define NEW_LIST_NODE(elm,tail,new_list){ \
  BPLONG tmp; \
  tmp = ADDTAG(heap_top,LST); \
  FOLLOW(heap_top++) = elm; \
  FOLLOW(heap_top++) = tail; \
  new_list = tmp; \
}

extern BPLONG no_gcs;

#define ALLOCATE_NEW_PAREA_BLOCK(size) { \
       BPLONG_PTR old_parea_low_addr; \
       num_parea_expansions++; \
       parea_updated = 1; \
       old_parea_low_addr = parea_low_addr; \
       BP_MALLOC(parea_low_addr,size,"parea"); \
       FOLLOW(parea_low_addr) = (BPLONG)old_parea_low_addr; \
       FOLLOW(parea_low_addr+1) = size; \
       parea_up_addr = parea_low_addr + size; \
       parea_water_mark =  parea_up_addr - SMALL_MARGIN; \
       curr_fence     = (CHAR_PTR)(parea_low_addr+2); \
}


