/* -*- c-basic-offset: 4 ; tab-width: 4 -*- */

#ifdef MPI

/*------------------------------------------------------------------------*/

#include "bprolog.h"
#include "prism.h"
#include "em.h"
#include "mp.h"
#include <mpi.h>

/*------------------------------------------------------------------------*/

/* flags.c */
extern int params_after_vbem;

/*------------------------------------------------------------------------*/

/* mic.c */
void    myquit(int, const char *);

/* em_aux.c */
void    release_occ_switches(void);
void    release_num_sw_vals (void);
void    release_occ_position(void);

/* em_aux_ml.c */
int     check_smooth(int *);

/* em_aux_vb.c */
int     check_smooth_vb(void);
void    get_param_means(void);

/* graph.c */
int     graph_stats(int *);

/* mp_em_ml.c */
void    mpm_share_preconds_em(int *);
void    mps_share_preconds_em(int *);
int     mpm_run_em(EM_ENG_PTR);
int     mps_run_em(EM_ENG_PTR);

/* mp_em_vb.c */
void    mpm_share_preconds_vbem(void);
void    mps_share_preconds_vbem(void);
int     mpm_run_vbem(VBEM_ENG_PTR);
int     mps_run_vbem(VBEM_ENG_PTR);

/* mp_em_aux.c */
void    release_sw_msg_buffers(void);

/*------------------------------------------------------------------------*/

int pc_mpm_prism_em(void)
{
    struct EM_Engine em_eng;

    /* [28 Aug 2007, by yuizumi]
     * occ_switches[] will be freed in pc_import_occ_switches/1.
     * occ_position[] is not allocated.
     */
    RETURN_ON_ERROR(check_smooth(&em_eng.smooth));
    mpm_share_preconds_em(&em_eng.smooth);
    RETURN_ON_ERROR(mpm_run_em(&em_eng));
    release_sw_msg_buffers();
    release_num_sw_vals();

    return
        bpx_unify(bpx_get_call_arg(1,6), bpx_build_integer(em_eng.iterate))  &&
        bpx_unify(bpx_get_call_arg(2,6), bpx_build_float(em_eng.lambda))     &&
        bpx_unify(bpx_get_call_arg(3,6), bpx_build_float(em_eng.likelihood)) &&
        bpx_unify(bpx_get_call_arg(4,6), bpx_build_float(em_eng.bic))        &&
        bpx_unify(bpx_get_call_arg(5,6), bpx_build_float(em_eng.cs))         &&
        bpx_unify(bpx_get_call_arg(6,6), bpx_build_integer(em_eng.smooth));
}

int pc_mps_prism_em(void)
{
    struct EM_Engine em_eng;

    mps_share_preconds_em(&em_eng.smooth);
    RETURN_ON_ERROR(mps_run_em(&em_eng));
    release_sw_msg_buffers();
    release_occ_switches();
    release_num_sw_vals();
    release_occ_position();

    return BP_TRUE;
}

int pc_mpm_prism_vbem(void)
{
    struct VBEM_Engine vb_eng;

    RETURN_ON_ERROR(check_smooth_vb());
    mpm_share_preconds_vbem();
    RETURN_ON_ERROR(mpm_run_vbem(&vb_eng));
    release_sw_msg_buffers();
    release_num_sw_vals();

    return
        bpx_unify(bpx_get_call_arg(1,2), bpx_build_integer(vb_eng.iterate)) &&
        bpx_unify(bpx_get_call_arg(2,2), bpx_build_float(vb_eng.free_energy));
}

int pc_mps_prism_vbem(void)
{
    struct VBEM_Engine vb_eng;

    mps_share_preconds_vbem();
    RETURN_ON_ERROR(mps_run_vbem(&vb_eng));
    release_sw_msg_buffers();
    release_occ_switches();
    release_num_sw_vals();
    release_occ_position();

    return BP_TRUE;
}

int pc_mpm_prism_both_em(void)
{
    struct EM_Engine   em_eng;
    struct VBEM_Engine vb_eng;

    RETURN_ON_ERROR(check_smooth_vb());
    mpm_share_preconds_vbem();
    RETURN_ON_ERROR(mpm_run_vbem(&vb_eng));

    em_eng.smooth      =  1 ;
    em_eng.lambda      = 0.0;
    em_eng.likelihood  = 0.0;
    em_eng.iterate     =  0 ;
    em_eng.bic         = 0.0;
    em_eng.cs          = 0.0;

    switch (params_after_vbem) {
    case VB_UPDATE_PARAM_MEAN:
        get_param_means();
        break;
    case VB_UPDATE_PARAM_MAX :
        RETURN_ON_ERROR(mpm_run_em(&em_eng));
        break;
    }

    release_sw_msg_buffers();
    release_num_sw_vals();

    return
        bpx_unify(bpx_get_call_arg(1,7), bpx_build_integer(em_eng.iterate))   &&
        bpx_unify(bpx_get_call_arg(2,7), bpx_build_integer(vb_eng.iterate))   &&
        bpx_unify(bpx_get_call_arg(3,7), bpx_build_float(em_eng.lambda))      &&
        bpx_unify(bpx_get_call_arg(4,7), bpx_build_float(em_eng.likelihood))  &&
        bpx_unify(bpx_get_call_arg(5,7), bpx_build_float(vb_eng.free_energy)) &&
        bpx_unify(bpx_get_call_arg(6,7), bpx_build_float(em_eng.bic))         &&
        bpx_unify(bpx_get_call_arg(7,7), bpx_build_float(em_eng.cs));
}

int pc_mps_prism_both_em(void)
{
    struct EM_Engine   em_eng;
    struct VBEM_Engine vb_eng;

    mps_share_preconds_vbem();
    RETURN_ON_ERROR(mps_run_vbem(&vb_eng));

    em_eng.smooth = 1;

    switch (params_after_vbem) {
    case VB_UPDATE_PARAM_MEAN:
        get_param_means();
        break;
    case VB_UPDATE_PARAM_MAX :
        RETURN_ON_ERROR(mps_run_em(&em_eng));
        break;
    }

    release_sw_msg_buffers();
    release_occ_switches();
    release_num_sw_vals();
    release_occ_position();

    return BP_TRUE;
}

/*------------------------------------------------------------------------*/

int pc_mpm_import_graph_stats(void)
{
    int dummy[4] = { 0 };
    int stats[4];
    double avg_shared;

    MPI_Reduce(dummy, stats, 4, MPI_INT, MPI_SUM, 0, MPI_COMM_WORLD);
    avg_shared = (double)(stats[3]) / stats[0];

    return
        bpx_unify(bpx_get_call_arg(1,4), bpx_build_integer(stats[0])) &&
        bpx_unify(bpx_get_call_arg(2,4), bpx_build_integer(stats[1])) &&
        bpx_unify(bpx_get_call_arg(3,4), bpx_build_integer(stats[2])) &&
        bpx_unify(bpx_get_call_arg(4,4), bpx_build_float(avg_shared));
}

int pc_mps_import_graph_stats(void)
{
    int dummy[4];
    int stats[4];

    graph_stats(stats);
    MPI_Reduce(stats, dummy, 4, MPI_INT, MPI_SUM, 0, MPI_COMM_WORLD);

    mp_debug("# subgoals     = %d", stats[0]);
    mp_debug("# goal nodes   = %d", stats[1]);
    mp_debug("# switch nodes = %d", stats[2]);
    mp_debug("# sharings     = %d", stats[3]);

    return BP_TRUE;
}

/*------------------------------------------------------------------------*/

#endif /* MPI */
